"""
Cloud mask inference pipeline for LOI-60 products.
Contact: YK
"""

from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

import numpy as np
import pandas as pd
import satio
import xarray as xr
from evotrain.models.loi.data.preprocessing import scale_s2_l2a_bands
from evotrain.models.loi.infer.feats_head import get_feats_head
from evotrain.models.loi.infer.load import load_model_files
from evotrain.models.loi.io import save_geotiff
from loguru import logger
from satio.collections import L2ACollection
from satio_pc import random_string
from scipy.ndimage import binary_erosion
from skimage.transform import resize

from veg_workflows.composite import cubify
from veg_workflows.paths.models import CLOUDSEN_MODELS_60M
from veg_workflows.products import Loi10Product


@cubify
def erode_nan_neighbors(data, iterations=1):
    """Erode NaN borders from a 2D array."""
    valid_mask = ~np.isnan(data)
    for _ in range(iterations):
        valid_mask = binary_erosion(valid_mask)
    data[~valid_mask] = np.nan
    return data


class TerrascopeCloudInferencePipeline:
    def __init__(
        self,
        product_id: str,
        model_name,
        model_idx,
        version="v999",
        loi_path: Path = Path("/vitodata/vegteam_vol2/products/"),
        loi_tag: str = "LOI-60",
        s2grid=None,
    ):
        self.model_name = model_name
        self.model_idx = model_idx
        self.version = version

        self._model = None
        self._class_mapping = None
        self._model_config = None

        self.product_id = product_id

        self.loi_path = loi_path
        self.loi_tag = loi_tag

        self._satio_collection = None

        self._s2grid = (
            s2grid if s2grid is not None else satio.layers.load("s2grid")
        )
        self.bounds, self.epsg = self.get_tile_bounds(self.tile, self._s2grid)

    @property
    def model_path(self):
        return f"{CLOUDSEN_MODELS_60M}/{self.model_name}/version_0/"

    @property
    def satio_collection(self) -> L2ACollection:
        if self._satio_collection is None:
            self._satio_collection = self._load_collection()
            self._satio_collection.df = self._satio_collection.df[
                self._satio_collection.df.product_id == self.product_id
            ]
        return self._satio_collection

    @property
    def tile(self):
        return self.product_id.split("_")[5][1:]

    def get_tile_bounds(self, tile, s2grid):
        row = s2grid[s2grid.tile == tile].iloc[0]
        return row.bounds, row.epsg

    @property
    def model(self):
        if self._model is None:
            self._model, self._class_mapping, self._model_config = (
                self._load_model()
            )
        return self._model

    @property
    def class_mapping(self):
        if self._class_mapping is None:
            _ = self.model  # Load the model
        return self._class_mapping

    @property
    def model_config(self):
        if self._model_config is None:
            _ = self.model
        return self._model_config

    @property
    def loi10prod(self):
        loi10prod = Loi10Product(
            s2_product_id=self.product_id,
            project_name="LCFM",
            version=self.version,
            products_base_path=self.loi_path,
            product_name=self.loi_tag,
        )
        return loi10prod

    @property
    def date(self):
        date_id = self.loi10prod.date_id.split("_")[0]
        date_id = datetime.strptime(date_id, "%Y%m%dT%H%M%S")
        date = date_id.strftime("%Y-%m-%d")
        return date

    @property
    def bands(self):
        bands_list = self.model_config["bands_config"]["s2_bands"]
        bands_list = [
            band[4:] if band.startswith("l2a-") else band
            for band in bands_list
        ]
        return bands_list

    def _load_model(self):
        """Load the model from the given path."""
        model, class_mapping, model_config = load_model_files(
            self.model_path, idx=self.model_idx, load_last=False
        )
        return model, class_mapping, model_config

    def _load_collection(
        self, collection_name="lcfm_10percent", s2grid=None
    ) -> L2ACollection:
        from veg_workflows.collections import VegSatioCollections

        veg_collections = VegSatioCollections(s2grid=s2grid)
        collection = getattr(veg_collections, collection_name)
        collection = collection.filter_bounds(self.bounds, self.epsg)
        return collection

    def get_raw_inputs(self) -> Tuple[np.ndarray, Dict[str, Any]]:
        """Get and prepare input data from a product ID.

        Args:
            product_id: Sentinel-2 product ID
            collection_name: Name of the collection

        Returns:
            Tuple containing the input array and metadata
        """
        logger.info("Getting raw inputs")
        product_collection = self.satio_collection

        # Get bands data

        darrs_dict = product_collection.load(self.bands, resolution=60)
        darr = xr.concat(
            [darrs_dict[band] for band in self.bands], dim="band"
        ).isel(timestamp=0)

        scl_20 = product_collection.load(["SCL"], resolution=20)["SCL"].isel(
            timestamp=0
        )
        scl_60_arr = resize(
            scl_20.data,
            (darr.y.size, darr.x.size),
            order=0,
            preserve_range=True,
        )
        return darr, scl_60_arr

    def preprocess(
        self,
        darr: xr.DataArray,
        scl_array: np.ndarray,
        fill_value: int = 10_000,
        interpolate_nans: bool = False,
        erode_nans: int = 0,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Preprocess the input data.

        Args:
            input_array: Raw input array
        Returns:
            Tuple of preprocessed data and features
        """
        logger.info("Preprocessing input data")
        nodata_mask = scl_array == 0
        darr = darr.astype(np.float32)
        darr = darr.where(~nodata_mask, np.nan)
        data_interp = darr.data

        if np.isnan(data_interp).any(axis=0).any():
            if erode_nans > 0:
                logger.debug(f"Eroding nans neighbors {erode_nans} times")
                data_interp = erode_nan_neighbors(
                    data_interp, iterations=erode_nans
                )
                nodata_mask = np.isnan(
                    data_interp[0]
                )  # update expanded nodata
            if interpolate_nans:
                from veg_workflows.composite import nan_interpolate

                logger.debug("Interpolating nans")
                data_interp = nan_interpolate(darr.data)
            else:
                logger.debug(f"Filling nans with {fill_value}")
                data_interp = darr.fillna(fill_value).data
        else:
            data_interp = darr.data

        # Scale data
        scaled_data = scale_s2_l2a_bands(
            data_interp,
            scaling_factor=10_000,
            apply_logistic=True,
            k_factor=self.model_config["scaling_config"]["k_factor"],
            apply_jitter=False,
            k_factor_jitter=0,
        )

        # Generate additional features
        feats_head = get_feats_head(
            scaled_data.shape,
            self.epsg,
            self.bounds,
            self.date,
            self.model_config,
        )

        return scaled_data, feats_head, nodata_mask

    def inference(
        self,
        preprocessed_data: np.ndarray,
        feats_head: np.ndarray,
        nodata_mask: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Run model inference on preprocessed data.

        Args:
            preprocessed_data: Preprocessed input data
            feats_head: Additional features
            export: Whether to export the results

        Returns:
            Tuple of probability outputs and class outputs (argmax)
        """
        logger.info("Starting model inference")
        model_output = self.model.predict(preprocessed_data, feats_head)
        out = model_output.squeeze()

        # Get argmax output
        out_argmax = np.argmax(out.copy(), axis=0).astype(np.float32)

        # Apply masks
        out[:, nodata_mask] = np.nan
        out_argmax[nodata_mask] = np.nan

        return out, out_argmax

    def export(
        self,
        probs: np.ndarray,
        probs_argmax: np.ndarray,
        nodata_value: int = 255,
        scale: float = 1 / 250,
        dst_path_probs: Optional[Path] = None,
        dst_path_mask: Optional[Path] = None,
    ):
        """Export results to GeoTIFF files."""
        logger.info("Exporting results")
        self._export_probs(
            probs,
            dst_path=dst_path_probs,
            nodata_value=nodata_value,
            scale=scale,
        )

        self._export_mask(
            probs_argmax,
            dst_path=dst_path_mask,
            nodata_value=nodata_value,
        )

    def _export_probs(
        self,
        probs,
        dst_path=None,
        nodata_value=255,
        scale=1 / 250,
    ):
        probs_nodata_mask = np.isnan(probs)
        probs = np.round(probs / scale).astype(np.uint8)
        probs[probs_nodata_mask] = nodata_value

        bands_names = [
            f"PROB_{k}"
            for k in self.model_config["labels_config"][
                "cloudsen12_mergedclouds_extrasnow"
            ].keys()
        ]
        tags = {
            "bands_names": bands_names,
            "legend": "0 Surface \n 1 Clouds \n2 Shadows \n3 Snow \n255 No data",
        }

        # Use custom path if provided, otherwise use default from loi10prod
        save_path_probs = dst_path or self.loi10prod.path("probs")
        save_path_probs.parent.mkdir(parents=True, exist_ok=True, mode=0o775)
        # Save probability output
        save_geotiff(
            probs,
            self.bounds,
            self.epsg,
            save_path_probs,  # Use the chosen path
            scales=[1 / 250] * probs.shape[0],
            nodata_value=nodata_value,
            bands_names=bands_names,
            tags=tags,
        )

        return save_path_probs

    def _export_mask(self, probs_argmax, dst_path=None, nodata_value=255):
        probs_argmax_nodata_mask = np.isnan(probs_argmax)
        probs_argmax[probs_argmax_nodata_mask] = nodata_value
        probs_argmax = probs_argmax.astype(np.uint8)

        save_path_mask = dst_path or self.loi10prod.path("mask")
        save_path_mask.parent.mkdir(parents=True, exist_ok=True, mode=0o775)

        bands_names = ["CLASS"]
        tags = {
            "band_names": ["CLASS"],
            "legend": "0 Surface \n 1 Clouds \n2 Shadows \n3 Snow \n255 No data",
        }
        # Save class output
        save_geotiff(
            probs_argmax.astype(np.uint8),
            self.bounds,
            self.epsg,
            save_path_mask,  # Use the chosen path
            nodata_value=nodata_value,
            bands_names=bands_names,
            tags=tags,
        )

        return save_path_mask

    def run(
        self,
        export=True,
        fill_value: int = 10_000,
        interpolate_nans: bool = False,
        erode_nans: int = 0,
    ):
        """Run the inference pipeline."""
        darr, scl_array = self.get_raw_inputs()
        preprocessed_data, feats_head, nodata_mask = self.preprocess(
            darr,
            scl_array,
            fill_value=fill_value,
            interpolate_nans=interpolate_nans,
            erode_nans=erode_nans,
        )
        probs, probs_argmax = self.inference(
            preprocessed_data, feats_head, nodata_mask
        )
        if export:
            self.export(probs, probs_argmax)
        return probs, probs_argmax


class AWSCloudInferencePipeline(TerrascopeCloudInferencePipeline):
    def __init__(
        self,
        product_id: str,
        model_name: str,
        model_idx: int,
        version="v999",
        loi_path: Path = Path("products/"),
        loi_tag: str = "LOI-60",
        output_bucket: str = "vito-lcfm",
        cache_folder: str = None,
        s2_db_uri: str = "s3://vito-lcfm/global_processing/collections/l2a_aws_catalog_v3.parquet",
        s2grid=None,
        s2_db=None,
    ):
        """Initialize the cloud inference model."""

        self.output_bucket = output_bucket
        if cache_folder is None:
            cache_folder = f".veg-workflows-cache-{random_string(8)}"
        self._cache_folder = Path(cache_folder)
        self._download_folder = self._cache_folder / f"l2a-{product_id}"

        self._s2_db_uri = s2_db_uri
        self._s2_db_product = None
        self._s2_db = s2_db

        super().__init__(
            product_id,
            model_name=model_name,
            model_idx=model_idx,
            version=version,
            loi_path=loi_path,
            loi_tag=loi_tag,
            s2grid=s2grid,
        )

    @property
    def model_path(self):
        return (
            "s3://vito-lcfm/global_processing/models/"
            f"{self.model_name}/version_0"
        )

    @property
    def s2_db(self):
        if self._s2_db is None:
            self._s2_db = pd.read_parquet(self._s2_db_uri)
        return self._s2_db

    @property
    def s2_db_product(self):
        if self._s2_db_product is None:
            self._s2_db_product = self.s2_db[
                self.s2_db.product_id == self.product_id
            ]
        return self._s2_db_product

    def _load_model(
        self,
    ):
        from evotrain.models.loi.infer.load import load_s3_model_files

        model, class_mapping, model_config = load_s3_model_files(
            self.model_path, idx=self.model_idx, load_last=False
        )
        return model, class_mapping, model_config

    def _download_product(self, download_folder):
        from veg_workflows.aws import AWSL2ADownloader

        s2_db = self.s2_db
        s2_db_product = s2_db[s2_db.product_id == self.product_id].iloc[0]
        tile = s2_db_product.tile

        l2a_downloader = AWSL2ADownloader()
        l2a_downloader.download_l2a(
            self.s2_db_product,
            bands_10m=[],
            bands_20m=["SCL"],
            bands_60m=self.bands,
            tile=tile,
            download_folder=self._download_folder,
        )

    def _load_collection(self) -> L2ACollection:
        self._download_product(self._download_folder)
        l2a_coll = L2ACollection.from_path(
            self._download_folder, s2grid=self._s2grid
        )
        l2a_coll = l2a_coll.filter_bounds(self.bounds, self.epsg)
        return l2a_coll

    def export(
        self,
        probs: np.ndarray,
        probs_argmax: np.ndarray,
        nodata_value: int = 255,
        scale: float = 1 / 250,
        dry_run: bool = False,
    ):
        """Export results to GeoTIFF files."""
        import os

        from elogs.s3 import S3BucketReader

        aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID")
        aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY")

        s3 = S3BucketReader.from_credentials(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            bucket=self.output_bucket,
            aws_region="eu-central-1",
        )

        dst_key = self.loi10prod.path("probs")
        dst_path_probs = self._cache_folder / dst_key.name

        self._export_probs(
            probs,
            dst_path=dst_path_probs,
            nodata_value=nodata_value,
            scale=scale,
        )

        if not dry_run:
            s3.upload(
                str(dst_path_probs),
                dst_key,
            )
        else:
            logger.info(f"{dst_path_probs} -> {dst_key}")

        dst_key = self.loi10prod.path("mask")
        dst_path_mask = self._cache_folder / dst_key.name
        self._export_mask(
            probs_argmax,
            dst_path=dst_path_mask,
            nodata_value=nodata_value,
        )

        if not dry_run:
            s3.upload(
                str(dst_path_mask),
                dst_key,
            )
        else:
            logger.info(f"{dst_path_mask} -> {dst_key}")

    def clean_up(self):
        import shutil

        shutil.rmtree(self._cache_folder, ignore_errors=True)

    def run(
        self,
        export=True,
        clean_up=True,
        dry_run=False,
        fill_value=10_000,
        interpolate_nans=False,
        erode_nans=0,
    ):
        """Run the inference pipeline."""
        darr, scl_array = self.get_raw_inputs()
        preprocessed_data, feats_head, nodata_mask = self.preprocess(
            darr,
            scl_array,
            fill_value=fill_value,
            interpolate_nans=interpolate_nans,
            erode_nans=erode_nans,
        )
        probs, probs_argmax = self.inference(
            preprocessed_data, feats_head, nodata_mask
        )
        if export:
            self.export(probs, probs_argmax, dry_run=dry_run)

        if clean_up:
            self.clean_up()
        return probs, probs_argmax


if __name__ == "__main__":
    # AWS example
    s2_db_uri = "s3://vito-lcfm/global_processing/collections/l2a_aws_catalog_v3.parquet"

    s2_db = pd.read_parquet(s2_db_uri)
    s2grid = satio.layers.load("s2grid")

    product_id = s2_db.sample(1, random_state=0).iloc[0].product_id

    # test_aws_production_loi-60_2501241011_v8aa_2_v100test
    model_name = "2501241011_v8aa"
    model_idx = 2
    s2_db.tile

    cloud_pipeline = AWSCloudInferencePipeline(
        product_id, model_name, model_idx, s2_db=s2_db, s2grid=s2grid
    )

    probs, mask = cloud_pipeline.run()

    # Terrascope example
    product_id = "S2A_MSIL2A_20200612T023601_N0500_R089_T50NKJ_20230327T190018"
    terrascope_pipeline = TerrascopeCloudInferencePipeline(
        product_id, model_name, model_idx, s2grid=s2grid
    )

    probs, mask = terrascope_pipeline.run()
